Package: TestGenerator
Type: Package
Title: Integration Unit Tests for Pharmacoepidemiological Studies
Version: 0.8.0
Authors@R: c(
    person("Cesar", "Barboza", , "c.barboza@darwin-eu.org", role = c("aut", "cre"), comment = c(ORCID = "0009-0002-4453-3071")),
    person("Ioanna", "Nika", , "i.nika@darwin-eu.org", role = "aut"),
    person("Ger", "Inberg", , "g.inberg@erasmusmc.nl", role = c("aut"), comment = c(ORCID = "0000-0001-8993-8748")),
    person("Adam", "Black", , "black@ohdsi.org", role = "aut", comment = c(ORCID = "0000-0001-5576-8701"))
  )
Maintainer: Cesar Barboza <c.barboza@darwin-eu.org>
Description: An R interface to load testing data in the 'OMOP' Common Data Model ('CDM'). An input file, csv or xlsx, can be converted to a 'CDMConnector' object. This object can be used to execute and test studies that use the 'CDM' <https://www.ohdsi.org/data-standardization/>.
License: Apache License (>= 2)
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 4.1.0)
Imports: jsonlite, readxl, readr, CDMConnector (>= 2.5.1), DBI, dplyr,
        checkmate, glue, duckdb, cli, rlang, withr, tibble, testthat,
        openxlsx, omopgenerics, RPostgres, fs, stringr, usethis
Suggests: knitr, rmarkdown, curl, httr, odbc, CohortConstructor
VignetteBuilder: knitr
Config/testthat/edition: 3
URL: https://github.com/darwin-eu/TestGenerator,
        https://darwin-eu.github.io/TestGenerator/
BugReports: https://github.com/darwin-eu/TestGenerator/issues
NeedsCompilation: no
Packaged: 2026-05-17 22:00:45 UTC; runner
Author: Cesar Barboza [aut, cre] (ORCID:
    <https://orcid.org/0009-0002-4453-3071>),
  Ioanna Nika [aut],
  Ger Inberg [aut] (ORCID: <https://orcid.org/0000-0001-8993-8748>),
  Adam Black [aut] (ORCID: <https://orcid.org/0000-0001-5576-8701>)
Repository: CRAN
Date/Publication: 2026-05-26 13:20:21 UTC
Built: R 4.6.0; ; 2026-05-26 16:08:50 UTC; unix

# Generated by roxygen2: do not edit by hand

export(cleanupTestCdm)
export(downloadTestData)
export(generateTestTables)
export(graphCohort)
export(patientsCDM)
export(readPatients)
export(readPatients.csv)
export(readPatients.xl)
export(useGithubAction)
import(cli)
import(dplyr)
importFrom(CDMConnector,cdmFromCon)
importFrom(CDMConnector,downloadEunomiaData)
importFrom(CDMConnector,eunomiaDir)
importFrom(DBI,dbAppendTable)
importFrom(DBI,dbConnect)
importFrom(DBI,dbDisconnect)
importFrom(DBI,dbExecute)
importFrom(checkmate,assert)
importFrom(checkmate,assertCharacter)
importFrom(checkmate,assertDirectoryExists)
importFrom(checkmate,assertFileExists)
importFrom(duckdb,duckdb)
importFrom(fs,file_exists)
importFrom(fs,path_file)
importFrom(glue,glue)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,toJSON)
importFrom(omopgenerics,insertTable)
importFrom(openxlsx,addWorksheet)
importFrom(openxlsx,createWorkbook)
importFrom(openxlsx,saveWorkbook)
importFrom(openxlsx,writeData)
importFrom(readr,read_csv)
importFrom(readxl,excel_sheets)
importFrom(readxl,read_excel)
importFrom(stringr,str_subset)
importFrom(testthat,test_path)
importFrom(tibble,tibble)
importFrom(tools,file_path_sans_ext)
importFrom(usethis,proj_path)
importFrom(utils,download.file)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# TestGenerator

<!-- badges: start -->

[![R-CMD-check](https://github.com/darwin-eu/TestGenerator/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/darwin-eu/TestGenerator/actions/workflows/R-CMD-check.yaml)
[![codecov](https://codecov.io/github/darwin-eu/TestGenerator/branch/main/graph/badge.svg)](https://app.codecov.io/github/darwin-eu/TestGenerator?branch=main)
[![CRAN
status](https://www.r-pkg.org/badges/version/TestGenerator)](https://CRAN.R-project.org/package=TestGenerator)
<!-- badges: end -->

Did my cohort pick the correct number of patients? Am I calculating an
intersection in the right way? Is that the expected value for treatment
duration? It only takes one incorrect parameter to get incoherent
results in a pharmacoepidemiological study, and testing calculations on
huge, complex databases is very challenging.

That is why TestGenerator is useful: it lets you push a small sample of
patients to unit test a study on the OMOP CDM. It includes tools to
create a blank CDM with a complete vocabulary and check whether the code
is doing what we expect in very specific cases.

This package is based on the unit tests written for the [Erasmus MC
Ranitidine
Study](https://github.com/mi-erasmusmc/RanitidineStudy/blob/master/unitTesting_README.md).

## Installation

Install the released version from CRAN:

``` r
install.packages("TestGenerator")
```

## Basic workflow

TestGenerator starts from a small patient dataset. The data can be
stored in an Excel workbook, with one sheet per OMOP CDM table, or in a
folder of CSV files, with one file per table.

For help creating an Excel input from scratch, see the [Start from a
Blank Excel
Template](https://darwin-eu.github.io/TestGenerator/articles/gettingStarted.html#start-from-a-blank-excel-template)
section of the website.

The package then converts those files into a Unit Test Definition JSON
file. This JSON file is the object you keep in your package tests.

``` r
TestGenerator::readPatients(
  filePath = "inst/extdata/icu_sample_population.xlsx",
  testName = "icu_sample",
  outputPath = "tests/testthat/testCases",
  cdmVersion = "5.4"
)
```

If `outputPath = NULL`, the JSON file is written to
`tests/testthat/testCases`, which is the usual location for package
tests.

You can also call the Excel and CSV readers directly:

``` r
TestGenerator::readPatients.xl(
  filePath = "inst/extdata/icu_sample_population.xlsx",
  testName = "icu_sample",
  outputPath = "tests/testthat/testCases",
  cdmVersion = "5.4"
)

TestGenerator::readPatients.csv(
  filePath = "inst/extdata/icu_sample_population_csv",
  testName = "icu_sample",
  outputPath = "tests/testthat/testCases",
  cdmVersion = "5.4",
  reduceLargeIds = FALSE
)
```

## Create a test CDM

Use `patientsCDM()` to load one Unit Test Definition into a blank OMOP
CDM. By default, this creates a local DuckDB CDM with the small patient
population and the vocabulary needed for testing.

``` r
cdm <- TestGenerator::patientsCDM(
  pathJson = "tests/testthat/testCases",
  testName = "icu_sample",
  cdmVersion = "5.4"
)
```

If `pathJson = NULL`, TestGenerator looks for the JSON file in
`tests/testthat/testCases`.

The example below uses the sample ICU population included in the
package.

``` r
file_path <- system.file(
  "extdata",
  "icu_sample_population.xlsx",
  package = "TestGenerator"
)

output_path <- file.path(tempdir(), "testgenerator-example")
dir.create(output_path, recursive = TRUE, showWarnings = FALSE)

TestGenerator::readPatients(
  filePath = file_path,
  testName = "icu_sample",
  outputPath = output_path,
  cdmVersion = "5.4"
)

cdm <- TestGenerator::patientsCDM(
  pathJson = output_path,
  testName = "icu_sample",
  cdmVersion = "5.4"
)

DBI::dbDisconnect(CDMConnector::cdmCon(cdm), shutdown = TRUE)
unlink(output_path, recursive = TRUE)
```

## Use it in a package test

The most useful pattern is to keep a small JSON test case in
`tests/testthat/testCases`, build a CDM inside a `testthat` test, run
your study code, and assert the expected result.

``` r
testthat::test_that("cohort construction returns the expected patients", {
  cdm <- TestGenerator::patientsCDM(
    pathJson = "tests/testthat/testCases",
    testName = "icu_sample",
    cdmVersion = "5.4"
  )
  withr::defer(
    DBI::dbDisconnect(CDMConnector::cdmCon(cdm), shutdown = TRUE)
  )

  cohort_set <- CDMConnector::readCohortSet(
    system.file("extdata", "test_cohorts", package = "TestGenerator")
  )

  cdm <- CDMConnector::generateCohortSet(
    cdm = cdm,
    cohortSet = cohort_set,
    name = "test_cohorts"
  )

  result <- cdm[["test_cohorts"]] |>
    dplyr::collect()

  testthat::expect_equal(
    sort(unique(result$subject_id)),
    c(1, 2, 4, 5, 6, 7)
  )
})
```

The exact expectation should come from the micro population you
designed. Good tests usually check subject counts, inclusion or
exclusion rules, cohort dates, or treatment durations that are easy to
verify by hand.

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## ----all-actions--------------------------------------------------------------
# library(TestGenerator)
# 
# useGithubAction(overwrite = TRUE)

## ----one-action---------------------------------------------------------------
# useGithubAction(dbms_type = "postgresql", overwrite = TRUE)
# useGithubAction(dbms_type = "sqlserver", overwrite = TRUE)
# useGithubAction(dbms_type = "databricks", overwrite = TRUE)

#' Read a parquet file using DuckDB.
#' @param file Path to the parquet file.
#' @return A data frame.
#' @noRd
read_parquet_file <- function(file) {
  con <- DBI::dbConnect(duckdb::duckdb())
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE), add = TRUE)
  path_sql <- gsub("'", "''", path.expand(file))
  DBI::dbGetQuery(con, paste0("SELECT * FROM read_parquet('", path_sql, "')"))
}

#' Generates an Excel file with sheets that correspond to an OMOP-CDM tables.
#'
#' @param tableNames A list specifying the table names to include in the Excel file.
#' @param cdmVersion The CDM version to use for creating the requested tables (either 5.3 or 5.4).
#' @param outputFolder The folder where the Excel file will be saved.
#' @param filename The name of the Excel file. It does not include the extension .xlsx.
#'
#' @return An Excel file with the tables requested.
#' @export
#' @importFrom openxlsx createWorkbook addWorksheet writeData saveWorkbook
generateTestTables <- function(
    tableNames = c("person", "observation_period", "visit_occurrence",
                   "visit_detail", "condition_occurrence", "drug_exposure",
                   "procedure_occurrence", "measurement", "observation", "death",
                   "drug_era", "condition_era", "dose_era", "location", "care_site",
                   "provider"),
    cdmVersion,
    outputFolder,
    filename = paste0(
      "test_cdm_",
      cdmVersion
      )
    ) {

  if(!(cdmVersion %in% c("5.3", "5.4"))){
    stop("Invalid cdm version should be 5.3 or 5.4")
  }

  tableNames <- tolower(tableNames)

  cdmSpecificationPath <- system.file("cdmTableSpecifications", paste0("emptycdm_", cdmVersion), package = "TestGenerator")
  # check table names
  fileNames <- tolower(tools::file_path_sans_ext(basename(list.files(path = cdmSpecificationPath, full.names = TRUE))))


  toExclude <- c("concept",
                 "concept_ancestor",
                 "concept_class",
                 "concept_cpt4",
                 "concept_relationship",
                 "concept_synonym",
                 "domain",
                 "drug_strength",
                 "relationship",
                 "vocabulary")

  fileNames <- setdiff(fileNames, toExclude)

  invalidTableNames <-setdiff(tableNames, fileNames)

  if (length(invalidTableNames) > 0) {
    stop(paste("The following filenames are invalid:", paste0(invalidTableNames, collapse = ", ")))
  }

  if (!dir.exists(outputFolder)) {
    dir.create(outputFolder, recursive = TRUE)
  }

  # output path for excel file
  excelFileOutPath <- file.path(outputFolder, paste0(filename, ".xlsx"))

  wb <- openxlsx::createWorkbook()

  for(tableName in tableNames){

    # path to parquet file
    file <- file.path(cdmSpecificationPath, paste0(tableName, ".parquet"))

    parquetFile <- read_parquet_file(file)

    openxlsx::addWorksheet(wb,  tableName)
    openxlsx::writeData(wb, sheet = tableName, parquetFile)
  }

  # save the workbook
  openxlsx::saveWorkbook(wb, excelFileOutPath, overwrite = TRUE)
}

test_that("Valid basic table names", {

  cdmVersion <- "5.4"
  outputFolder <- file.path(tempdir(), "testOutputExcel")
  dir.create(outputFolder)

  expect_no_error(
    generateTestTables(
      cdmVersion = cdmVersion,
      outputFolder = outputFolder
    )
    )

  unlink(outputFolder, recursive = TRUE)

})

test_that("invalid table names", {

  tableNames <- c("table_that_does_not_exist", "vocabulary", "visit_occurrence")
  invalidTableNames <- c("table_that_does_not_exist", "vocabulary")
  cdmVersion <- "5.4"
  outputFolder <- file.path(tempdir(), "testOutputExcel")
  dir.create(outputFolder)

  expect_error(generateTestTables(tableNames = tableNames,
                     cdmVersion = cdmVersion,
                     outputFolder = outputFolder
                     ), paste("The following filenames are invalid:", paste0(invalidTableNames, collapse = ", ")))

  unlink(outputFolder, recursive = TRUE)

})


test_that("invalid cdm version", {

  tableNames <- c("visit_occurrence")
  cdmVersion <- "10.1"
  outputFolder <- file.path(tempdir(), "testOutputExcel")
  dir.create(outputFolder)

  expect_error(generateTestTables(tableNames = tableNames,
                                               cdmVersion = cdmVersion,
                                               outputFolder = outputFolder
  ), "Invalid cdm version should be 5.3 or 5.4")

  unlink(outputFolder, recursive = TRUE)

})


test_that("output is generated with correct specifications, lower or uppercase naming should not matter", {

  tableNames <-   tableNames <- c("person", "observation_period", "visit_occurrence",
                                  "visit_detail", "condition_occurrence", "drug_exposure",
                                  "procedure_occurrence", "measurement", "observation", "death",
                                  "drug_era", "condition_era", "dose_era", "location", "care_site",
                                  "provider", "COST")
  cdmVersion <- "5.4"
  outputFolder <- file.path(tempdir(), "testOutputExcel")
  dir.create(outputFolder)

  generateTestTables(tableNames = tableNames,
                     cdmVersion = cdmVersion,
                     outputFolder = outputFolder
                     )

  excelOutputPath <- file.path(outputFolder, paste0("test_cdm_", cdmVersion, ".xlsx"))

  wb <- openxlsx::loadWorkbook(excelOutputPath)
  sheetNames <- openxlsx::getSheetNames(excelOutputPath)

  sheetData <- data.frame()
  sheetData$cost <- openxlsx::read.xlsx(wb, sheet = "cost")
  sheetData$visit_occurrence <- openxlsx::read.xlsx(wb, sheet = "visit_occurrence")

  expect_true(all(tolower(tableNames) %in% sheetNames))

  parquetCostFilePath <- system.file("cdmTableSpecifications", paste0("emptycdm_", cdmVersion), "cost.parquet", package = "TestGenerator")
  parquetCost <- TestGenerator:::read_parquet_file(parquetCostFilePath)
  visitOccurrenceFilePath <- system.file("cdmTableSpecifications", paste0("emptycdm_", cdmVersion), "visit_occurrence.parquet", package = "TestGenerator")
  parquetVisitOccurrence <- TestGenerator:::read_parquet_file(visitOccurrenceFilePath)

  expect_identical(colnames(parquetCost), colnames(sheetData$cost))
  expect_identical(colnames(parquetVisitOccurrence), colnames(sheetData$visit_occurrence))
  unlink(outputFolder, recursive = TRUE)
})

#' Download Test Data Files
#'
#' @param datasetName The data set name as found on https://github.com/darwin-eu/EunomiaDatasets. The
#'  data set name corresponds to the folder with the data set ZIP files
#' @param cdmVersion The OMOP CDM version. This version will appear in the suffix of the data file,
#'  for example: synpuf_5.3.zip. Default: '5.3'
#' @param pathToData The path where the Eunomia data is stored on the file system., By default the
#'  value of the environment variable "EUNOMIA_DATA_FOLDER" is used.
#' @param overwrite Control whether the existing archive file will be overwritten should it already exist.
#' @return
#' Invisibly returns the destination if the download was successful.
#'
#' @importFrom utils download.file
#'
#' @examples
#' \donttest{
#' downloadTestData(pathToData = tempdir())
#' }
#' @export
downloadTestData <- function(datasetName = "mimicIV",
                             cdmVersion = "5.3",
                             pathToData = Sys.getenv("STUDY_DATASETS"),
                             overwrite = FALSE) {
  checkmate::assertChoice(datasetName, choices = studyDatasets())
  if (cdmVersion != "5.3") {
    rlang::abort("Only CDM v5.3 is supported currently!")
  }
  if (is.null(pathToData) || is.na(pathToData) || pathToData == "") {
    stop("The pathToData argument must be specified. Consider setting the TEST_DATASETS environment variable, for example in the .Renviron file.")
  }
  if (!dir.exists(pathToData)) {
    dir.create(pathToData, recursive = TRUE)
  }
  zipName <- glue::glue("{datasetName}.zip")
  if (file.exists(file.path(pathToData, zipName)) && !overwrite) {
    rlang::inform(glue::glue("Dataset already exists ({file.path(pathToData, zipName)})\n      Specify `overwrite = TRUE` to overwrite existing zip archive"))
    return(invisible(pathToData))
  }

  pathToFile <- NULL
  remoteResource <- "https://physionet.org/static/published-projects/mimic-iv-demo-omop/mimic-iv-demo-data-in-the-omop-common-data-model-0.9.zip"
  if (!is.null(checkRemoteFileAvailable(remoteResource))) {
    pb <- cli::cli_progress_bar(format = "[:bar] :percent :elapsed",
                                type = "download")
    withr::with_options(list(timeout = 5000), {
      utils::download.file(url = remoteResource,
                           destfile = file.path(pathToData, zipName), mode = "wb",
                           method = "auto", quiet = FALSE, extra = list(progressfunction = function(downloaded, total) {
                             progress <- min(1, downloaded/total)
                             cli::cli_progress_update(id = pb, set = progress)
                           }))
    })
    cli::cli_progress_done(id = pb)
    pathToFile <- file.path(pathToData, zipName)
  }
  return(pathToFile)
}

studyDatasets <- function() {
  c("mimicIV")
}

#' Check if a given remote file is available for download
#'
#' @param remoteFile a remote resource
#' @return NULL if the remote resource is not available, other "success"
checkRemoteFileAvailable <- function(remoteFile) {
  try_GET <- function(x, ...) {
    tryCatch(
      httr::GET(url = x, httr::timeout(100), ...),
      error = function(e) conditionMessage(e),
      warning = function(w) conditionMessage(w)
    )
  }
  is_response <- function(x) {
    class(x) == "response"
  }

  # First check internet connection
  if (!curl::has_internet()) {
    message("No internet connection.")
    return(NULL)
  }
  # Then try for timeout problems
  resp <- try_GET(remoteFile)
  if (!is_response(resp)) {
    message(resp)
    return(NULL)
  }
  # Then stop if status > 400
  if (httr::http_error(resp)) {
    httr::message_for_status(resp)
    return(NULL)
  }
  return("success")
}

#' Converts a sample of patients into Unit Testing Definition JSON file.
#'
#' @param filePath Path to the test patient data in Excel format. The Excel has sheets that represent tables from the OMOP-CDM, e.g. person, drug_exposure, condition_ocurrence, etc.
#' @param testName A name of the test population in character.
#' @param outputPath Path of the output file, if NULL, a folder will be created in the project folder inst/testCases.
#' @param cdmVersion cdm version, default "5.4".
#' @param extraTable Name of non-standard tables to be included in the test CDM.
#'
#' @return A JSON file with sample patients inside the project directory.
#'
#' @importFrom readxl read_excel excel_sheets
#' @importFrom jsonlite toJSON
#' @importFrom checkmate assertDirectoryExists assertCharacter assertFileExists assert
#' @importFrom glue glue
#' @import cli
#'
#' @examples
#' filePath <- system.file("extdata", "testPatientsRSV.xlsx", package = "TestGenerator")
#' readPatients(filePath = filePath, outputPath = tempdir())
#'
#' @export
readPatients <- function(filePath = NULL,
                         testName = "test",
                         outputPath = NULL,
                         cdmVersion = "5.4",
                         extraTable = FALSE) {

  checkmate::assertFileExists(filePath)
  fileExtension <- tools::file_ext(filePath)
  checkmate::assertTRUE(fileExtension %in% c("csv", "xlsx"))

  if (fileExtension == "csv") {
    readPatients.csv(filePath = filePath,
                     testName = testName,
                     outputPath = outputPath,
                     cdmVersion = cdmVersion)
  } else {
    readPatients.xl(filePath = filePath,
                    testName = testName,
                    outputPath = outputPath,
                    cdmVersion = cdmVersion,
                    extraTable = extraTable)
  }
}

#' Converts a sample of patients in XLSX format into Unit Testing Definition JSON file.
#'
#' @param filePath Path to the test patient data in Excel format. The Excel has sheets that represent tables from the OMOP-CDM, e.g. person, drug_exposure, condition_ocurrence, etc.
#' @param testName A name of the test population in character.
#' @param outputPath Path to write the test JSON files. If NULL, the files will be written at the project's testthat folder, i.e. tests/testthat/testCases.
#' @param cdmVersion cdm version, default "5.4".
#' @param extraTable TRUE or FALSE. If TRUE, non-standard tables will be included in the test CDM.
#'
#' @return A directory with the test JSON files with sample patients inside the project directory.
#'
#' @importFrom readxl read_excel excel_sheets
#' @importFrom jsonlite toJSON
#' @importFrom checkmate assertDirectoryExists assertCharacter assertFileExists assert
#' @importFrom testthat test_path
#' @importFrom glue glue
#' @import cli
#'
#' @examples
#' filePath <- system.file("extdata", "testPatientsRSV.xlsx", package = "TestGenerator")
#' readPatients.xl(filePath = filePath, outputPath = tempdir())
#'
#' @export
readPatients.xl <- function(filePath = NULL,
                            testName = "test",
                            outputPath = NULL,
                            cdmVersion = "5.4",
                            extraTable = FALSE) {

  checkmate::assertCharacter(filePath)
  checkmate::assertFileExists(filePath)

  # Check columns

  cdmTables <- checkTablesColumns(cdmVersion, filePath, extraTable)


  # Convert to JSON
  testCaseFile <- jsonlite::toJSON(cdmTables,
                                   dataframe = "rows",
                                   pretty = TRUE)

  # Create testPath folder
  testPath <- createOutputFolder(outputPath, testName)

  # Write file
  write(testCaseFile, file = testPath)
  if (checkmate::checkFileExists(testPath)) {
    cli::cli_alert_success(glue::glue("Unit Test Definition Created Successfully: '{testName}'"))
  } else {
    cli::cli_alert_danger("Unit Test Definition Creation Failed")
    stop()
  }
}

#' Converts a sample of patients in CSV format into a Unit Testing Definition JSON file.
#'
#' @param filePath Path to the test patient data in CSV format. Multiple CSV files representing tables tables from the OMOP-CDM must be provided, e.g. person.csv, drug_exposure.csv, condition_ocurrence.csv, etc.
#' @param testName Name for the test population file in character.
#' @param outputPath Path of the output file, if NULL, a folder will be created in the project folder inst/testCases.
#' @param cdmVersion cdm version, default "5.4".
#' @param reduceLargeIds Reduces the length of very long ids generally in int64 format, such as those found in the MIMIC-IV database.
#'
#' @return A JSON file with sample patients inside the project directory.
#'
#' @importFrom readr read_csv
#' @importFrom jsonlite toJSON
#' @importFrom checkmate assertDirectoryExists assertCharacter assertFileExists assert
#' @importFrom glue glue
#' @importFrom tools file_path_sans_ext
#' @import cli
#'
#' @examples
#' filePath <- system.file("extdata", "mimic_sample", package = "TestGenerator")
#' readPatients.csv(
#'   filePath = filePath,
#'   outputPath = tempdir(),
#'   cdmVersion = "5.3"
#'   )
#'
#' @export
readPatients.csv <- function(filePath = NULL,
                             testName = "test",
                             outputPath = NULL,
                             cdmVersion = "5.4",
                             reduceLargeIds = FALSE) {

  checkmate::assertDirectoryExists(filePath)
  checkmate::assertCharacter(cdmVersion)
  checkmate::assertTRUE(cdmVersion %in% c("5.3", "5.4"))

  # Check column
  cdmTables <- fileColumnCheck(filePath, cdmVersion)

  if (reduceLargeIds) {
    cdmTables <- convertIds(cdmTables)
  }

  # Convert to JSON
  testCaseFile <- jsonlite::toJSON(cdmTables,
                                   dataframe = "rows",
                                   pretty = TRUE)

  # Create testPath folder
  testPath <- createOutputFolder(outputPath, testName)

  # Write file
  write(testCaseFile, file = testPath)
  if (checkmate::checkFileExists(testPath)) {
    cli::cli_alert_success(glue::glue("Unit Test Definition Created Successfully: '{testName}'"))
  } else {
    cli::cli_alert_danger("Unit Test Definition Creation Failed")
    stop()
  }
}

#' Check if the given tables and columns are valid and return the loaded data.
#'
#' @param cdmVersion cdm version
#' @param filePath Path to the test patient data in xlsx format
#' @param extraTable if extra tables are provided or not, default FALSE
#'
#' @return a named list containing the loaded data
checkTablesColumns <- function(cdmVersion, filePath, extraTable = FALSE) {
  checkmate::assertFileExists(filePath)

  patientTables <- readxl::excel_sheets(filePath)

  expectedTables <- spec_cdm_field[[cdmVersion]] %>%
    dplyr::pull(cdmTableName) %>%
    unique()

  if (extraTable) {
    if (!all(patientTables %in% unique(expectedTables))) {
      nonStandardTables <- setdiff(tolower(patientTables), expectedTables)
      cli::cli_alert_success(glue::glue("All tables are valid. Non-standard table(s) in test data: {glue::glue_collapse(nonStandardTables, sep = ', ', last = ' and ')}"))
    }
  } else {
    invalidTables <- setdiff(tolower(patientTables), expectedTables)
    if (invalidTables %>% length() > 0) {
      cli::cli_alert_danger(glue::glue("The following tables are invalid: {glue::glue_collapse(invalidTables, sep = ', ', last = ' and ')}"))
      stop()
    } else {
      cli::cli_alert_success(glue::glue("All tables are valid"))
    }
  }
  cdmTables <- lapply(patientTables,
                      readxl::read_excel,
                      path = filePath,
                      .name_repair = tolower) # cast column names to lowercase
  names(cdmTables) <- tolower(patientTables)

  return(cdmTables)
}

fileColumnCheck <- function(filePath, cdmVersion) {
  checkmate::assertDirectoryExists(filePath)
  checkmate::assertCharacter(cdmVersion)
  checkmate::assertTRUE(cdmVersion %in% c("5.3", "5.4"))
  csvFiles <- list.files(filePath, pattern = ".csv", full.names = TRUE)
  csvFilesNames <- list.files(filePath, pattern = ".csv")
  checkmate::assertCharacter(csvFiles, any.missing = FALSE, min.len = 1)
  checkmate::assertCharacter(csvFilesNames, any.missing = FALSE, min.len = 1)
  currentTables <- spec_cdm_field[[cdmVersion]] %>%
    dplyr::pull(cdmTableName) %>%
    unique()
  patientTables <- list()
  report <- list()
  for (i in 1:length(csvFiles)) {
    tableName <- tools::file_path_sans_ext(csvFilesNames[i])
    if (tableName %in% currentTables) {
      cdmTable <- readr::read_csv(csvFiles[i], show_col_types = FALSE)
      if (nrow(cdmTable) != 0) {
        names(cdmTable) <- tolower(names(cdmTable))
        currentCoulumns <- names(cdmTable)
        expectedColumns <- spec_cdm_field[[cdmVersion]] %>%
          dplyr::filter(cdmTableName == tableName) %>%
          dplyr::pull(cdmFieldName)
        expectedColumns <- gsub("\"", "", expectedColumns)
        if (all(currentCoulumns %in% expectedColumns)) {
          patientTables[[tableName]] <- cdmTable
        } else {
          cli::cli_alert_danger(glue::glue("'{tableName}' table columns do not match"))
          stop()
        }
      } else {
        report[["empty"]] <- append(report[["empty"]], glue::glue("{tableName}"))
      }
    }
  }
  if (!is.null(report[['empty']])) {
    empty <- paste(report[['empty']], collapse = ", ")
    cli::cli_alert_warning("Empty Tables Found:")
    cli::cli_text(empty)
  }
  names(patientTables) <- tolower(names(patientTables))
  return(patientTables)
}

convertIds <- function(cdmTables) {
  report <- list()
  for (tables in names(cdmTables)) {
    # tables <- "vocabulary"
    for (columns in names(cdmTables[[tables]])) {
      # columns <- "vocabulary_concept_id"
      if (columns %in% c("person_id",
                         "care_site_id",
                         "condition_era_id",
                         "condition_occurrence_id",
                         "device_exposure_id",
                         "visit_occurrence_id",
                         "dose_era_id",
                         "drug_era_id",
                         "drug_exposure_id",
                         "fact_id_1",
                         "fact_id_2",
                         "measurement_id",
                         "observation_id",
                         "observation_period_id",
                         "procedure_occurrence_id",
                         "specimen_id",
                         "visit_detail_id",
                         "preceding_visit_detail_id",
                         "preceding_visit_occurrence_id",
                         "vocabulary_concept_id")) {

        uniqueIdValues <- unique(cdmTables[[tables]][[columns]])
        idValues <- abs(cdmTables[[tables]][[columns]]) %>%
          abs() %>%
          format(scientific = FALSE, trim = TRUE) %>%
          substr(1, 9) %>%
          as.numeric()

        if (length(unique(idValues)) != length(unique(uniqueIdValues))) {
          if(!tables %in% c("person_id", "visit_occurrence_id", "condition_occurrence_id")) {
            cdmTables[[tables]][[columns]] <- seq(1, length(uniqueIdValues))
            report[["notUnique"]] <- append(report[["notUnique"]], glue::glue("{tables}"))
            # message(glue::glue("'{tables}' table with '{columns}' ids are not unique"))
            # message(glue::glue("'{tables}' table filled out with sequence of numbers"))
          } else {
            cli::cli_alert_danger(glue::glue("'{tables}' table with '{columns}' ids are not unique and couldn't fill with num sequence"))
            stop()
          }
        } else {
          cdmTables[[tables]][[columns]] <- idValues
          # report[["reduced"]] <- append(report[["reduced"]], glue::glue("{tables}"))
          # message(glue::glue("'{tables}' table and '{columns}' ids reduced succesfully"))
        }
      }
    }
  }
  if (!is.null(report[['notUnique']])) {
    notUnique <- paste(report[['notUnique']], collapse = ", ")
    cli::cli_alert_warning("Table with non unique ids and filled with num seq:")
    cli::cli_text(notUnique)
  }
  cli::cli_alert_success("IDs successfully reduced")
  # reduced <- paste(report[['reduced']], collapse = ", ")
  # cli::cli_text(reduced)
  return(cdmTables)
}

createOutputFolder <- function(outputPath, testName) {
  if (is.null(outputPath)) {
    testFolder <- testthat::test_path("testCases")
    if (!dir.exists(testFolder)) {
      dir.create(testFolder)
      testPath <- paste0(testFolder, "/", testName, ".json")
    } else {
      testPath <- paste0(testFolder, "/", testName, ".json")
    }
  } else {
    checkmate::assertCharacter(outputPath)
    checkmate::assertDirectoryExists(outputPath)
    testPath <- paste0(outputPath, "/", testName, ".json")
  }
  return(testPath)
}

#' Pushes test population into a blank CDM.
#'
#' @param pathJson Directory where the sample populations in json are located. If NULL, gets the default inst/testCases directory.
#' @param testName Name of the sample population JSON file. If NULL it will push the first sample population in the testCases directory.
#' @param cdmVersion cdm version, default "5.4".
#' @param cdmName Name of the cdm, default NULL.
#' @param dbms Database management system to use. One of "duckdb", "databricks",
#'   "sqlserver" or "postgresql". Default is "duckdb" which creates a local
#'   DuckDB CDM. For remote databases the function creates the CDM locally,
#'   trims the vocabulary, uploads to a new schema on the remote database,
#'   and returns the remote CDM reference. Remote databases require
#'   environment variables to be set. Call \code{usethis::edit_r_environ()}
#'   to set them.
#' @param writeSchema Optional schema name to use on the remote database.
#'   If NULL (default), a unique schema is created automatically. Only used
#'   when \code{dbms} is not "duckdb".
#'
#' @return A CDM reference object with a sample population.
#' @import dplyr cli
#' @importFrom DBI dbConnect dbAppendTable dbDisconnect dbExecute
#' @importFrom duckdb duckdb
#' @importFrom jsonlite fromJSON
#' @importFrom CDMConnector downloadEunomiaData eunomiaDir cdmFromCon
#' @importFrom omopgenerics insertTable
#'
#' @examples
#' \donttest{
#' filePath <- system.file("extdata", "testPatientsRSV.xlsx", package = "TestGenerator")
#' TestGenerator::readPatients(filePath = filePath, outputPath = tempdir())
#' cdm <- TestGenerator::patientsCDM(pathJson = tempdir(), testName = "test")
#' duckdb::duckdb_shutdown(duckdb::duckdb())
#' }
#' @export
patientsCDM <- function(pathJson = NULL,
                        testName = NULL,
                        cdmVersion = "5.4",
                        cdmName = NULL,
                        dbms = "duckdb",
                        writeSchema = NULL) {

  dbms <- match.arg(dbms, c("duckdb", "databricks", "sqlserver", "postgresql"))

  if (dbms != "duckdb") {
    .check_remote_env_vars(dbms)
    return(.patientsCDM_remote(
      pathJson = pathJson,
      testName = testName,
      cdmVersion = cdmVersion,
      cdmName = cdmName,
      dbms = dbms,
      writeSchema = writeSchema
    ))
  }

  if (is.null(pathJson)) {
    outputFolder <- testthat::test_path("testCases")
    if (dir.exists(outputFolder)) {
      pathJson <- outputFolder
    } else {
      cli::cli_alert_danger("testCases not found")
      stop()
    }
  }

  checkmate::assertClass(pathJson, "character")
  checkmate::assertDirectoryExists(pathJson)

  if (identical(list.files(pathJson), character(0))) {
    cli::cli_alert_danger("Directory empty. Provide Unit Test Definitions")
    stop()
  }

  testFiles <- list.files(pathJson, pattern = ".json")

  if (is.null(testName)) {
    testName <- testFiles[1]
  } else {
    checkmate::checkClass(testName, "character")
    testName <- paste0(testName, ".json")
  }

  fileName <- file.path(pathJson, testName)
  checkmate::assertFileExists(fileName)

  # Folder to download empty CDM

  if (!dir.exists(Sys.getenv("EUNOMIA_DATA_FOLDER"))) {
    Sys.setenv(EUNOMIA_DATA_FOLDER = tempdir())
  }

  # Check/Download vocabulary

  vocabPath <- file.path(Sys.getenv("EUNOMIA_DATA_FOLDER"),
                         glue::glue("empty_cdm_{cdmVersion}.zip"))

  if (!file.exists(vocabPath)) {
    CDMConnector::downloadEunomiaData(datasetName = "empty_cdm",
                                      cdmVersion = cdmVersion,
                                      pathToData = Sys.getenv("EUNOMIA_DATA_FOLDER"),
                                      overwrite = TRUE)
  }

  conn <- DBI::dbConnect(duckdb::duckdb(CDMConnector::eunomiaDir(datasetName = "empty_cdm",
                                                                 cdmVersion = cdmVersion)))
  cdm <- CDMConnector::cdmFromCon(con = conn,
                                  cdmSchema = "main",
                                  writeSchema = "main",
                                  cdmName = cdmName,
                                  cdmVersion = cdmVersion)

  # Read the JSON file into R
  jsonData <- jsonlite::fromJSON(fileName)

  # Check for the expected tables in the CDM
  expectedTables <- spec_cdm_field[[cdmVersion]] %>%
    dplyr::pull(cdmTableName) %>%
    unique()

  currentTables <- names(jsonData)

  nonStandardTables <- setdiff(currentTables, expectedTables)

  if (length(nonStandardTables) > 0) {
    cli::cli_alert_danger(glue::glue("Non-standard table(s) in test data: {glue::glue_collapse(nonStandardTables, sep = ', ', last = ' and ')}"))
  }
  standardTables <- setdiff(currentTables, nonStandardTables)
  cli::cli_alert_success(glue::glue("Standard table(s) in test data: {glue::glue_collapse(standardTables, sep = ', ', last = ' and ')}"))


  # Check for the expected columns in the CDM
  for (tableName in standardTables) {
    classTable <- class(jsonData[[tableName]])
    if (classTable == "data.frame") {
      currentCoulumns <- names(jsonData[[tableName]])
      expectedColumns <- spec_cdm_field[[cdmVersion]] %>%
        dplyr::filter(cdmTableName == tableName) %>%
        dplyr::pull(cdmFieldName)
      jsonData[[tableName]] <- jsonData[[tableName]] %>%
        select(currentCoulumns[currentCoulumns %in% expectedColumns])
      patientData <- as.data.frame(jsonData[[tableName]])
      DBI::dbAppendTable(conn, tableName, patientData)
    }
  }

  for (tableName in nonStandardTables) {
    patientData <- as.data.frame(jsonData[[tableName]])
    cdm <- omopgenerics::insertTable(cdm,
                                     tableName,
                                     patientData,
                                     overwrite = TRUE,
                                     temporary = FALSE)
  }

  cli::cli_alert_success("Patients pushed to blank CDM successfully")
  return(cdm)
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## ----postgresql-example-------------------------------------------------------
# library(TestGenerator)
# 
# cdm <- patientsCDM(
#   pathJson = "tests/testthat/testCases",
#   testName = "my_test_population",
#   cdmVersion = "5.4",
#   dbms = "postgresql"
# )

## ----write-schema-------------------------------------------------------------
# cdm <- patientsCDM(
#   pathJson = "tests/testthat/testCases",
#   testName = "my_test_population",
#   cdmVersion = "5.4",
#   dbms = "sqlserver",
#   writeSchema = "testgenerator_my_case"
# )

## ----cleanup-remote-----------------------------------------------------------
# cleanupTestCdm(cdm)

## ----test-pattern-------------------------------------------------------------
# testthat::test_that("study logic works on PostgreSQL", {
#   cdm <- NULL
#   on.exit({
#     if (!is.null(cdm)) {
#       TestGenerator::cleanupTestCdm(cdm)
#     }
#   }, add = TRUE)
# 
#   cdm <- TestGenerator::patientsCDM(
#     pathJson = "tests/testthat/testCases",
#     testName = "my_test_population",
#     cdmVersion = "5.4",
#     dbms = "postgresql"
#   )
# 
#   result <- myPackage::runMyStudy(cdm)
# 
#   testthat::expect_equal(result$n_subjects, 3)
# })

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## ----create-json--------------------------------------------------------------
# library(TestGenerator)
# 
# file_path <- system.file(
#   "extdata",
#   "icu_sample_population.xlsx",
#   package = "TestGenerator"
# )
# 
# output_path <- file.path(tempdir(), "testgenerator-example")
# dir.create(output_path, showWarnings = FALSE, recursive = TRUE)
# 
# readPatients(
#   filePath = file_path,
#   testName = "icu_sample",
#   outputPath = output_path,
#   cdmVersion = "5.4"
# )

## ----load-cdm-----------------------------------------------------------------
# cdm <- patientsCDM(
#   pathJson = output_path,
#   testName = "icu_sample",
#   cdmVersion = "5.4"
# )
# 
# cdm[["person"]]

## ----default-test-path--------------------------------------------------------
# cdm <- patientsCDM(
#   pathJson = NULL,
#   testName = "icu_sample",
#   cdmVersion = "5.4"
# )

## ----cohort-test--------------------------------------------------------------
# library(CDMConnector)
# library(dplyr)
# library(testthat)
# 
# test_cohorts <- system.file(
#   "extdata",
#   "test_cohorts",
#   package = "TestGenerator"
# )
# 
# cohort_set <- readCohortSet(test_cohorts)
# 
# cdm <- generateCohortSet(
#   cdm = cdm,
#   cohortSet = cohort_set,
#   name = "test_cohorts"
# )
# 
# cohort_attrition <- attrition(cdm[["test_cohorts"]])
# 
# excluded_records <- cohort_attrition |>
#   pull(excluded_records) |>
#   sum()
# 
# expect_equal(excluded_records, 0)

## ----generate-template--------------------------------------------------------
# generateTestTables(
#   tableNames = c(
#     "person",
#     "observation_period",
#     "visit_occurrence",
#     "condition_occurrence",
#     "drug_exposure",
#     "measurement"
#   ),
#   cdmVersion = "5.4",
#   outputFolder = output_path,
#   filename = "my_test_population"
# )

## ----csv-input----------------------------------------------------------------
# csv_path <- system.file(
#   "extdata",
#   "mimic_sample",
#   package = "TestGenerator"
# )
# 
# readPatients.csv(
#   filePath = csv_path,
#   testName = "mimic_sample",
#   outputPath = output_path,
#   cdmVersion = "5.4"
# )

## ----cleanup------------------------------------------------------------------
# DBI::dbDisconnect(CDMConnector::cdmCon(cdm), shutdown = TRUE)
# unlink(output_path, recursive = TRUE)

test_that("checkTablesColumns function works csv with test data", {
  filePath <- testthat::test_path("test_cdm_data.xlsx")
  checkmate::assertCharacter(filePath)
  checkmate::assertFileExists(filePath)
  cdmVersion <- "5.4"
  listPatientTables <- checkTablesColumns(cdmVersion, filePath, extraTable = FALSE)

  expect_in(names(listPatientTables), c("person", "observation_period", "condition_occurrence", "pregnancy",
                                        "measurement", "drug_exposure", "visit_detail", "visit_occurrence",
                                        "procedure_occurrence", "device_exposure", "observation", "note",
                                        "death", "note_nlp", "specimen", "fact_relationship"))
})

test_that("checkTablesColumns function works csv with test data - some tables and columns in uppercase", {
  filePath <- testthat::test_path("test_cdm_data_uppercase.xlsx")
  checkmate::assertCharacter(filePath)
  checkmate::assertFileExists(filePath)
  cdmVersion <- "5.4"
  listPatientTables <- checkTablesColumns(cdmVersion, filePath, extraTable = FALSE)

  expect_in(names(listPatientTables), c("person", "observation_period", "condition_occurrence", "drug_exposure"))

  lapply(names(listPatientTables), FUN = function(tableName) {
    expect_true(identical(colnames(listPatientTables[[tableName]]), tolower(colnames(listPatientTables[[tableName]]))))
  })
})

test_that("checkTablesColumns function works csv with table 'pregnancy'", {
  filePath <- testthat::test_path("test_cdm_data_pregnancy.xlsx")
  checkmate::assertCharacter(filePath)
  checkmate::assertFileExists(filePath)
  cdmVersion <- "5.4"
  listPatientTables <- checkTablesColumns(cdmVersion, filePath, extraTable = TRUE)

  expect_in(names(listPatientTables), c("person", "observation_period", "condition_occurrence", "pregnancy",
                                        "measurement", "drug_exposure", "visit_detail", "visit_occurrence",
                                        "procedure_occurrence", "device_exposure", "observation", "note",
                                        "death", "note_nlp", "specimen", "fact_relationship"))
})

test_that("checkTablesColumns function works csv with table 'pregnancy'", {
  filePath <- testthat::test_path("test_cdm_data_pregnancy.xlsx")
  checkmate::assertCharacter(filePath)
  checkmate::assertFileExists(filePath)
  cdmVersion <- "5.4"

  expect_error(checkTablesColumns(cdmVersion, filePath, extraTable = FALSE))
})

test_that("checkTablesColumns function works csv", {
  filePath <- testthat::test_path("test_cdm_data_pregnancy.xlsx")
  checkmate::assertCharacter(filePath)
  checkmate::assertFileExists(filePath)
  cdmVersion <- "5.4"
  expect_error(checkTablesColumns(cdmVersion, filePath, extraTable = "pregnancys"))
})

test_that("fileColumnCheck function works", {
  filePath <- testthat::test_path("mimic_sample")
  cdmVersion <- "5.3"
  listPatientTables <- fileColumnCheck(filePath, cdmVersion)
  expect_equal(names(listPatientTables), c("condition_occurrence",
                                           "drug_exposure",
                                           "measurement",
                                           "observation_period",
                                           "person",
                                           "visit_detail",
                                           "visit_occurrence"))
})

test_that("Reading patients XLSX and JSON creation for pregnancy", {
  filePath <- testthat::test_path("testPatientsRSV.xlsx")
  # outputPath <- file.path(tempdir(), "test1")
  # dir.create(outputPath)

  # outputPath explicitly NULL to create the testCases in the testthat folder
  readPatients.xl(filePath = filePath, outputPath = NULL, extraTable = FALSE)
  expect_true(file.exists(file.path(testthat::test_path("testCases"), "test.json")))
})

test_that("Reading patients XLSX and JSON creation", {
  filePath <- testthat::test_path("test_cdm_data_pregnancy.xlsx")
  # outputPath <- file.path(tempdir(), "test1")
  # dir.create(outputPath)

  # outputPath explicitly NULL to create the testCases in the testthat folder
  readPatients.xl(filePath = filePath, testName = "pregnancy", outputPath = NULL, extraTable = TRUE)
  expect_true(file.exists(file.path(testthat::test_path("testCases"), "pregnancy.json")))
})

test_that("Patients to CDM xlsx function", {
  skip_on_cran()
  filePath <- testthat::test_path("testPatientsRSV.xlsx")
  TestGenerator::readPatients.xl(filePath = filePath, outputPath = NULL)
  cdm <- TestGenerator::patientsCDM(pathJson = NULL, testName = "pregnancy")
  expect_equal(class(cdm), "cdm_reference")
  number_persons <- cdm[["person"]] %>% dplyr::pull(person_id)
  expect_equal(length(number_persons), 18)
  duckdb::duckdb_shutdown(duckdb::duckdb())
})

test_that("Patients to CDM xlsx function pregnancy extra table", {
  skip_on_cran()
  filePath <- testthat::test_path("test_cdm_data_pregnancy.xlsx")
  TestGenerator::readPatients.xl(filePath = filePath, testName = "pregnancy", outputPath = NULL, extraTable = TRUE)
  cdm <- TestGenerator::patientsCDM(pathJson = NULL, testName = "pregnancy")
  expect_equal(class(cdm), "cdm_reference")
  expect_equal(cdm$pregnancy %>% colnames(), c("pregnancy_occurrence_id", "person_id", "pregnancy_concept_id", "pregnancy_start_date"))
  number_persons <- cdm[["person"]] %>% dplyr::pull(person_id)
  expect_equal(length(number_persons), 18)
  duckdb::duckdb_shutdown(duckdb::duckdb())
})

test_that("Read patients empty tables xl", {
  skip_on_cran()
  filePath <- testthat::test_path("test_cdm_data.xlsx")
  TestGenerator::readPatients.xl(filePath = filePath, outputPath = NULL)
  cdm <- TestGenerator::patientsCDM(pathJson = NULL, testName = "test")
  expect_equal(class(cdm), "cdm_reference")
  number_persons <- cdm[["person"]] %>% dplyr::pull(person_id)
  expect_equal(length(number_persons), 18)
  duckdb::duckdb_shutdown(duckdb::duckdb())
})

test_that("Read patients empty xl", {
  skip_on_cran()
  filePath <- testthat::test_path("test_cdm_data.xlsx")
  TestGenerator::readPatients.xl(filePath = filePath, outputPath = NULL)
  cdm <- TestGenerator::patientsCDM(pathJson = NULL, testName = "test")
  expect_equal(class(cdm), "cdm_reference")
  number_persons <- cdm[["person"]] %>% dplyr::pull(person_id)
  expect_equal(length(number_persons), 18)
  duckdb::duckdb_shutdown(duckdb::duckdb())
})

test_that("Reading sample MIMIC patients CSV files and JSON creation", {
  filePath <- testthat::test_path("mimic_sample")
  outputPath <- testthat::test_path("testCases")
  # outputPath <- file.path(tempdir(), "test1")
  # dir.create(outputPath)
  readPatients.csv(
    filePath = filePath,
    testName = "mimic_sample",
    outputPath = NULL,
    cdmVersion = "5.3"
    )
  expect_true(file.exists(file.path(outputPath, "mimic_sample.json")))
  # unlink(outputPath, recursive = TRUE)
})

test_that("Reading MIMIC patients CSV files and JSON creation", {
  skip_on_cran()
  pathToData <- tempdir()
  pathToZipFile <- downloadTestData(datasetName = "mimicIV",
                                    cdmVersion = "5.3",
                                    pathToData = pathToData,
                                    overwrite = TRUE)
  if (!is.null(pathToZipFile)) {
    unzip(pathToZipFile, exdir = pathToData)
    filePath <- file.path(pathToData,
                          "mimic-iv-demo-data-in-the-omop-common-data-model-0.9",
                          "1_omop_data_csv")
    outputPath <- file.path(tempdir(), "test1")
    dir.create(outputPath)
    testName <- "test"
    cdmVersion <- "5.3"
    readPatients.csv(filePath = filePath,
                     testName = testName,
                     outputPath = outputPath,
                     cdmVersion = cdmVersion)
    expect_true(file.exists(file.path(outputPath, "test.json")))
    unlink(outputPath, recursive = TRUE)
  }
})

test_that("Mimic data Patients to CDM function", {
  skip_on_cran()
  pathToData <- tempdir()
  cdmVersion <- "5.3"
  pathToZipFile <- downloadTestData(datasetName = "mimicIV",
                                    cdmVersion = cdmVersion,
                                    pathToData = pathToData,
                                    overwrite = TRUE)
  if (!is.null(pathToZipFile)) {
    unzip(pathToZipFile, exdir = pathToData)
    filePath <- file.path(pathToData,
                          "mimic-iv-demo-data-in-the-omop-common-data-model-0.9",
                          "1_omop_data_csv")
    outputPath <- file.path(tempdir(), "test1")
    dir.create(outputPath)
    testName <- "test"
    readPatients.csv(filePath = filePath,
                     testName = testName,
                     outputPath = outputPath,
                     cdmVersion = cdmVersion,
                     reduceLargeIds = TRUE)
    cdmName <- "myCDM"
    cdm <- TestGenerator::patientsCDM(pathJson = outputPath, testName = "test", cdmName = cdmName)
    expect_equal(class(cdm), "cdm_reference")
    number_persons <- cdm[["person"]] %>% dplyr::pull(person_id) %>% length()
    expect_equal(number_persons, 100)
    expect_equal(CDMConnector::cdmName(cdm), cdmName)
    unlink(outputPath, recursive = TRUE)
    duckdb::duckdb_shutdown(duckdb::duckdb())
  }
})

test_that("convert ids function", {
  pathToData <- tempdir()
  cdmVersion <- "5.3"
  pathToZipFile <- downloadTestData(datasetName = "mimicIV",
                                    cdmVersion = cdmVersion,
                                    pathToData = pathToData,
                                    overwrite = TRUE)
  if (!is.null(pathToZipFile)) {
    unzip(pathToZipFile, exdir = pathToData)
    filePath <- file.path(pathToData,
                          "mimic-iv-demo-data-in-the-omop-common-data-model-0.9",
                          "1_omop_data_csv")
    cdmTables <- fileColumnCheck(filePath, cdmVersion)
    cdmTables <- convertIds(cdmTables)
    measurement_ids <- cdmTables$measurement %>% pull(measurement_id)
    expect_equal(measurement_ids, seq(1, length(measurement_ids)))
    unlink(pathToData, recursive = TRUE)
    unlink(filePath, recursive = TRUE)
    duckdb::duckdb_shutdown(duckdb::duckdb())
  }
})

test_that("Patients to CDM sqlserver and check cdm reference has attributes", {
  skip_on_cran()
  skip_if(Sys.getenv("SQLSERVER_SERVER") == "")
  skip_if(Sys.getenv("SQLSERVER_DBNAME") == "")
  skip_if(Sys.getenv("SQLSERVER_USER") == "")
  skip_if(Sys.getenv("SQLSERVER_PASSWORD") == "")
  skip_if(Sys.getenv("SQLSERVER_PORT") == "")
  cdm <- NULL
  on.exit({
    if (!is.null(cdm)) {
      TestGenerator::cleanupTestCdm(cdm)
    }
  }, add = TRUE)
  cdmVersion <- "5.4"
  filePath <- testthat::test_path(
    "test_cdm_data_pregnancy.xlsx"
    )
  TestGenerator::readPatients(
    filePath = filePath,
    testName = "pregnancy",
    outputPath = NULL,
    extraTable = TRUE
  )
  # Test no error in connection
  expect_no_error(
    cdm <- TestGenerator::patientsCDM(
      pathJson = NULL,
      testName = "pregnancy",
      cdmVersion = cdmVersion,
      dbms = "sqlserver",
      cdmName = "cdm"
    )
  )

  attr(cdm, "cdm_schema") |>
    stringr::str_detect(
      "testgenerator"
      ) |>
    expect_true()

  expect_no_error({
    cdm$cohort <- CohortConstructor::conceptCohort(
      cdm = cdm,
      conceptSet = list(a = 44500984L),
      name = "cohort")
  })

  cdm$cohort |>
    dplyr::collect() |>
    nrow() |>
    expect_equal(2)

})

test_that("Patients to CDM version 5.4", {
  skip_on_cran()
  cdmVersion <- "5.4"
  filePath <- testthat::test_path("test_cdm_data_pregnancy.xlsx")
  TestGenerator::readPatients(filePath = filePath, testName = "pregnancy", outputPath = NULL, extraTable = TRUE)
  cdm <- TestGenerator::patientsCDM(pathJson = NULL, testName = "pregnancy", cdmVersion = cdmVersion)
  expect_equal(class(cdm), "cdm_reference")
  expect_equal(CDMConnector::snapshot(cdm) %>% dplyr::pull("cdm_version"), cdmVersion)
  duckdb::duckdb_shutdown(duckdb::duckdb())
})

#' Deprecated cohort timeline visualisation
#'
#' @description
#' `graphCohort()` is deprecated and will be removed in a future release.
#'
#' @param subject_id Deprecated.
#' @param cohorts Deprecated.
#'
#' @return Invisibly returns `NULL`.
#' @export
graphCohort <- function(subject_id, cohorts = list()) {
  .Deprecated(
    msg = "`graphCohort()` is deprecated and will be removed in a future release."
  )
  invisible(NULL)
}

emptyCDM <- function(conn, cdm) {

  for (table_name in names(cdm)) {
    if (table_name %in% c("person",
                          "observation_period",
                          "visit_occurrence",
                          "visit_detail",
                          "condition_occurrence",
                          "drug_exposure",
                          "procedure_occurrence",
                          "device_exposure",
                          "measurement",
                          "observation",
                          "death",
                          "note",
                          "note_nlp",
                          "specimen",
                          "fact_relationship",
                          "location",
                          "care_site",
                          "provider",
                          "payer_plan_period",
                          "cost",
                          "drug_era",
                          "dose_era",
                          "condition_era",
                          "metadata",
                          "cdm_source")) {
      DBI::dbExecute(conn, glue::glue("TRUNCATE TABLE {table_name}"))
    }
  }
  return(cdm)
}

test_that("CDM emptied", {
  skip("manual test")
  conn <- DBI::dbConnect(duckdb::duckdb(CDMConnector::eunomiaDir("synthea-allergies-10k")))
  cdm <- CDMConnector::cdmFromCon(conn, cdmSchema = "main", writeSchema = "main")
  cdm <- emptyCDM(cdm = cdm, con = conn)
  expect_equal(class(cdm), "cdm_reference")
  countPerson <- DBI::dbGetQuery(conn, "SELECT COUNT(*) FROM person")
  expect_equal(countPerson[[1]], 0)
  DBI::dbDisconnect(conn, shutdown = TRUE)
})

#' Create GitHub Action workflows for specific DBMS
#'
#' Copies GitHub Action workflow templates from the package's
#' \code{inst/workflows} directory to the \code{.github/workflows} directory of
#' the current project. The copied workflows can be used to run backend-specific
#' tests for PostgreSQL, SQL Server, and Databricks.
#'
#' @param dbms_type A character vector of supported DBMS types to process.
#'   Supported values are \code{"postgresql"}, \code{"sqlserver"}, and
#'   \code{"databricks"}.
#' @param overwrite A logical value indicating whether to overwrite existing
#'   workflow files in the \code{.github/workflows} directory. Defaults to
#'   \code{FALSE}.
#'
#' @return Invisibly returns \code{TRUE} when the selected workflow files are
#'   copied successfully.
#'
#' @importFrom checkmate assertCharacter assertDirectoryExists
#' @importFrom fs path_file
#' @importFrom stringr str_subset
#' @importFrom usethis proj_path
#'
#' @examples
#' \dontrun{
#' useGithubAction(dbms_type = "postgresql")
#' useGithubAction(
#'   dbms_type = c("postgresql", "sqlserver", "databricks"),
#'   overwrite = TRUE
#' )
#' }
#'
#' @export
useGithubAction <- function(
  dbms_type = c(
    "postgresql",
    "sqlserver",
    "databricks"
  ),
  overwrite = FALSE
) {

  check_supported_dbms(dbms_type)
  origin_workflows_folder <- system.file(
    "workflows",
    package = "TestGenerator"
  )
  checkmate::assertDirectoryExists(
    origin_workflows_folder
    )
  origin_file_names <- list.files(
    origin_workflows_folder,
    full.names = TRUE
  )
  destination_workflows_dir <- usethis::proj_path(
    ".github",
    "workflows"
  )
  create_dir(destination_workflows_dir)

  for (i in seq_along(dbms_type)) {
    origin_path <- stringr::str_subset(
      origin_file_names,
      dbms_type[i]
    )
    destination_path <- file.path(
      destination_workflows_dir,
      fs::path_file(origin_path)
      )
    checkmate::assertDirectoryExists(destination_workflows_dir)
    copy_status <- file.copy(
      origin_path,
      destination_path,
      overwrite = overwrite
    )
    if (isTRUE(copy_status) & isTRUE(file.exists(destination_path))) {
      cli::cli_alert_success(
        glue::glue(
          "{fs::path_file(origin_path)} inserted to {destination_workflows_dir}"
        )
      )
      } else if (isFALSE(copy_status) & isTRUE(file.exists(destination_path))) {
        stop(
          cli::cli_alert_danger(
            glue::glue(
              "{fs::path_file(origin_path)} workflow already exists in {destination_workflows_dir}"
              )
            )
        )
      } else if (isTRUE(copy_status) & isFALSE(file.exists(destination_path))) {
        stop(
          cli::cli_alert_danger(
            glue::glue(
              "{fs::path_file(origin_path)} was copied but not to {destination_workflows_dir}"
            )
          )
        )
      }
  }
  invisible(TRUE)
}

create_dir <- function(path) {
  if (dir.exists(path)) {
    return(invisible(FALSE))
  }
  else if (file_exists(path)) {
    stop(
      glue::glue(
        "{path} exists but is not a directory"
        )
      )
  }
  dir.create(
    path,
    recursive = TRUE
    )
  cli::cli_alert_success(
    glue::glue(
      "Creating {path}"
      )
  )
  invisible(TRUE)
}

check_supported_dbms <- function(dbms_type) {
  checkmate::assertCharacter(dbms_type)
  for (i in 1:length(dbms_type)) {
    if (!dbms_type[i] %in% c("postgresql", "sqlserver", "databricks")) {
      stop(
        cli::cli_alert_danger(
          glue::glue(
            "{dbms_type[i]} is not a supported DBMS"
            )
        )
      )
    }
    }
  invisible(TRUE)
}

test_that("workflow files created correctly in .github/workflows", {

  test_project <- file.path(
    tempdir(),
    "testgenerator-github-action-test"
  )
  unlink(
    test_project,
    recursive = TRUE
  )
  withr::defer(
    unlink(
      test_project,
      recursive = TRUE
    ),
    envir = testthat::teardown_env()
  )
  dir.create(
    test_project,
    recursive = TRUE
  )
  writeLines(
    c(
      "Package: testgeneratorActionTest",
      "Title: Test Project",
      "Version: 0.0.0.9000",
      "Description: Temporary project for TestGenerator tests.",
      "License: MIT",
      "Encoding: UTF-8"
    ),
    file.path(test_project, "DESCRIPTION")
  )
  usethis::local_project(
    test_project,
    .local_envir = testthat::teardown_env()
  )

  workflows_folder <- system.file(
    "workflows",
    package = "TestGenerator"
  )

  expect_true(
    dir.exists(workflows_folder)
    )

  dbms_types <- c(
    "postgresql",
    "sqlserver",
    "databricks"
    )

  workflow_files <- list.files(
    workflows_folder
    )

  github_folder <- usethis::proj_path(
    ".github",
    "workflows"
    )

  # Test individual workflow file creation
  for (i in 1:length(dbms_types)) {

    workflow_file_name <- stringr::str_subset(
      workflow_files,
      dbms_types[i]
    )

    destination_path <- file.path(
      github_folder,
      workflow_file_name
    )

    unlink(
      destination_path,
      recursive = TRUE
    )

    expect_false(
      checkmate::testFileExists(
        destination_path
      )
    )

    useGithubAction(
      dbms_type = dbms_types[i],
      overwrite = TRUE
      )

    expect_true(
      file.exists(destination_path)
      )

    unlink(
      destination_path,
      recursive = TRUE
    )

  }

  # Test all files were created at once
  useGithubAction()

  for (i in 1:length(workflow_files)) {

    path <- file.path(
      github_folder,
      workflow_files[i]
    )

    expect_true(
      file.exists(path)
    )

  }

  # Expect error if overwrite is FALSE
  expect_error(
    useGithubAction(
      overwrite = FALSE
      )
  )

  # Expect error if incorrect DBMS type
  expect_error(
    useGithubAction(
      dbms_type = c("sqlserver", "spark")
    )
  )

})

# Internal: create test CDM on a remote database
.patientsCDM_remote <- function(pathJson, testName, cdmVersion, cdmName, dbms, writeSchema = NULL) {

  n_steps <- 5L

  # ---- Step 1: Create local DuckDB CDM ----
  cli::cli_progress_step("Step 1/{n_steps}: Creating local DuckDB test CDM")
  local_cdm <- patientsCDM(
    pathJson = pathJson,
    testName = testName,
    cdmVersion = cdmVersion,
    cdmName = cdmName,
    dbms = "duckdb"
  )

  # Track the duckdb file path so we can clean it up later
  local_con <- CDMConnector::cdmCon(local_cdm)
  duckdb_path <- local_con@driver@dbdir

  # ---- Step 2: Trim vocabulary ----
  cli::cli_progress_step("Step 2/{n_steps}: Trimming vocabulary")
  local_cdm <- CDMConnector::cdmTrimVocabulary(local_cdm)

  # ---- Step 3: Connect to remote database and create schema ----
  remote_con <- .connect_remote(dbms)
  if (!is.null(writeSchema)) {
    cli::cli_progress_step("Step 3/{n_steps}: Connecting to {dbms} using schema {.val {writeSchema}}")
    test_schema <- writeSchema
  } else {
    cli::cli_progress_step("Step 3/{n_steps}: Connecting to {dbms} and creating test schema")
    test_schema <- .create_test_schema(remote_con, dbms)
    cli::cli_alert_info("Created test schema: {.val {paste(test_schema, collapse = '.')}}")
  }

  # ---- Step 4: Upload CDM to remote database ----
  cli::cli_progress_step("Step 4/{n_steps}: Uploading test CDM to {dbms}")
  cdm_remote <- CDMConnector::copyCdmTo(
    con = remote_con,
    cdm = local_cdm,
    schema = test_schema,
    overwrite = TRUE
  )

  # ---- Step 5: Create CDM referemce ----

  cdm <- CDMConnector::cdmFromCon(
    con = remote_con,
    cdmSchema = test_schema,
    writeSchema = test_schema,
    cdmName = cdmName,
    cdmVersion = cdmVersion
  )

  # ---- Step 6: Clean up local DuckDB resources ----
  cli::cli_progress_step("Step 5/{n_steps}: Cleaning up local DuckDB files")
  DBI::dbDisconnect(local_con, shutdown = TRUE)

  # Remove duckdb file and any WAL files
  if (nzchar(duckdb_path) && file.exists(duckdb_path)) {
    unlink(duckdb_path)
    unlink(paste0(duckdb_path, ".wal"))
  }

  cli::cli_alert_success("Remote test CDM ready on {dbms}")
  return(cdm)
}

# Internal: early check that required env vars are set before doing any work
.check_remote_env_vars <- function(dbms) {
  required_vars <- switch(dbms,
                          "sqlserver" = c(
                            "SQLSERVER_SERVER",
                            "SQLSERVER_DBNAME",
                            "SQLSERVER_PORT",
                            "SQLSERVER_USER",
                            "SQLSERVER_PASSWORD"
                          ),
                          "postgresql" = c(
                            "POSTGRESQL_SERVER",
                            "POSTGRESQL_DBNAME",
                            "POSTGRESQL_PORT",
                            "POSTGRESQL_USER",
                            "POSTGRESQL_PASSWORD"
                          ),
                          "databricks" = c(
                            "DATABRICKS_HOST",
                            "DATABRICKS_TOKEN",
                            "DATABRICKS_HTTPPATH"
                          )
  )
  .check_env_vars(required_vars, dbms)
}

# Internal: connect to a remote database using environment variables
.connect_remote <- function(dbms) {
  switch(dbms,
         "sqlserver" = .connect_sqlserver(),
         "databricks" = .connect_databricks(),
         "postgresql" = .connect_postgresql(),
         cli::cli_abort("Unsupported dbms: {.val {dbms}}")
  )
}

.connect_sqlserver <- function() {
  required_vars <- c(
    "SQLSERVER_SERVER",
    "SQLSERVER_DBNAME",
    "SQLSERVER_PORT",
    "SQLSERVER_USER",
    "SQLSERVER_PASSWORD"
  )
  .check_env_vars(required_vars, "SQL Server")

  DBI::dbConnect(
    odbc::odbc(),
    Driver   = Sys.getenv("SQL_SERVER_DRIVER", "ODBC Driver 18 for SQL Server"),
    Server   = Sys.getenv("SQLSERVER_SERVER"),
    Database = Sys.getenv("SQLSERVER_DBNAME"),
    UID      = Sys.getenv("SQLSERVER_USER"),
    PWD      = Sys.getenv("SQLSERVER_PASSWORD"),
    TrustServerCertificate = "yes",
    Port     = Sys.getenv("SQLSERVER_PORT")
  )
}

.connect_postgresql <- function() {
  required_vars <- c(
    "POSTGRESQL_SERVER",
    "POSTGRESQL_DBNAME",
    "POSTGRESQL_PORT",
    "POSTGRESQL_USER",
    "POSTGRESQL_PASSWORD"
  )
  .check_env_vars(required_vars, "Postgresql")

  portStr <- trimws(Sys.getenv("POSTGRESQL_PORT"))
  if (!grepl("^[0-9]+$", portStr)) {
    cli::cli_abort(
      "{.envvar POSTGRESQL_PORT} must be a valid integer for PostgreSQL."
    )
  }
  port <- as.integer(portStr)

  DBI::dbConnect(
    RPostgres::Postgres(),
    host     = Sys.getenv("POSTGRESQL_SERVER"),
    dbname   = Sys.getenv("POSTGRESQL_DBNAME"),
    port     = port,
    user     = Sys.getenv("POSTGRESQL_USER"),
    password = Sys.getenv("POSTGRESQL_PASSWORD")
  )
}

.connect_databricks <- function() {
  required_vars <- c(
    "DATABRICKS_HOST",
    "DATABRICKS_TOKEN",
    "DATABRICKS_HTTPPATH",
    "DATABRICKS_USER"
  )
  .check_env_vars(required_vars, "Databricks")

  DBI::dbConnect(
    drv = odbc::odbc(),
    Driver = "Databricks ODBC Driver",
    Host = sub("^https://", "", Sys.getenv("DATABRICKS_HOST")),
    Port = 443,
    SSL = 1,
    ThriftTransport = 2,
    SparkServerType = 3,
    AuthMech = 3,
    HTTPPath = Sys.getenv("DATABRICKS_HTTPPATH"),
    UID = Sys.getenv("DATABRICKS_USER", "token"),
    PWD = Sys.getenv("DATABRICKS_TOKEN"),
    UseNativeQuery = 0,
    bigint = "numeric"
  )
}

# Internal: check that required environment variables are set
.check_env_vars <- function(vars, label) {
  missing <- vars[!nzchar(Sys.getenv(vars))]
  if (length(missing) > 0) {
    bullets <- stats::setNames(missing, rep("*", length(missing)))
    cli::cli_abort(c(
      "Missing required environment variables for {label}:",
      bullets,
      "i" = "Set them in your {.file .Renviron} by running {.code usethis::edit_r_environ()}"
    ))
  }
}

# Internal: create a unique test schema on the remote database
.create_test_schema <- function(con, dbms) {
  schema_name <- paste0("cdm_testgenerator_", format(Sys.time(), "%Y%m%d_%H%M%S"), "_", sample(1000:9999, 1))

  switch(dbms,
         "sqlserver" = {
           DBI::dbExecute(con, paste("CREATE SCHEMA", schema_name))
           schema_name
         },
         "postgresql" = {
           DBI::dbExecute(con, paste("CREATE SCHEMA", schema_name))
           schema_name
         },
         "databricks" = {
           catalog <- Sys.getenv("DATABRICKS_WORKSPACE", "hive_metastore")
           schema <- c(catalog = catalog, schema = schema_name)
           DBI::dbExecute(con, paste0(
             "CREATE SCHEMA IF NOT EXISTS ",
             paste(schema, collapse = ".")
           ))
           schema
         }
  )
}

#' Clean up a test CDM on a remote database
#'
#' Drops the schema containing the test CDM and disconnects from the database.
#'
#' @param cdm A CDM reference object created by \code{patientsCDM()} with a
#'   remote database backend.
#'
#' @return Invisibly returns NULL.
#' @export
cleanupTestCdm <- function(cdm) {
  con <- CDMConnector::cdmCon(cdm)
  schema <- CDMConnector::cdmWriteSchema(cdm)
  writeSchema <- paste(schema, collapse = ".")
  dbms_class <- class(con)

  if (!grepl("testgenerator", writeSchema, ignore.case = TRUE)) {
    cli::cli_abort("Refusing to drop schema {.val {writeSchema}} because it does not contain {.val testgenerator} in the name.")
  }

  if (any(grepl("OdbcConnection", dbms_class)) || any(grepl("Microsoft SQL Server", dbms_class))) {
    # SQL Server: drop all tables first, then drop schema
    tables <- CDMConnector::listTables(con, schema = schema)
    schema_name <- if (length(schema) == 1) schema else schema[length(schema)]
    for (tbl in tables) {
      tryCatch(
        DBI::dbExecute(con, paste0("DROP TABLE IF EXISTS ", schema_name, ".", tbl)),
        error = function(e) NULL
      )
    }
    DBI::dbExecute(con, paste("DROP SCHEMA", schema_name))
    cli::cli_alert_success("Dropped SQL Server schema: {.val {schema_name}}")
  } else if (any(grepl("PqConnection", dbms_class))) {
    # Postgresql: drop all tables first, then drop schema
    tables <- CDMConnector::listTables(con, schema = schema)
    schema_name <- if (length(schema) == 1) schema else schema[length(schema)]
    for (tbl in tables) {
      tryCatch(
        DBI::dbExecute(con, paste0("DROP TABLE IF EXISTS ", schema_name, ".", tbl)),
        error = function(e) NULL
      )
    }
    DBI::dbExecute(con, paste("DROP SCHEMA", schema_name))
    cli::cli_alert_success("Dropped Postgresql schema: {.val {schema_name}}")
  } else if (any(grepl("Spark|databricks", dbms_class, ignore.case = TRUE))) {
    # Databricks/Spark
    schema_full <- paste(schema, collapse = ".")
    DBI::dbExecute(con, paste0("DROP SCHEMA IF EXISTS ", schema_full, " CASCADE"))
    cli::cli_alert_success("Dropped Spark schema: {.val {schema_full}}")
  } else {
    cli::cli_abort("Could not detect database type for cleanup. Connection class: {.val {paste(dbms_class, collapse = ', ')}}")
  }

  DBI::dbDisconnect(con)
  cli::cli_alert_success("Disconnected from remote database")
  invisible(NULL)
}

---
title: "GitHub Actions"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{GitHub Actions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

TestGenerator can create GitHub Actions workflows for backend-specific package
tests. The goal is to make continuous integration check that your package works
against PostgreSQL, SQL Server, and Databricks/Spark, not only against a local
DuckDB test CDM.

## Create the workflow files

Create all supported workflows with:

```{r all-actions}
library(TestGenerator)

useGithubAction(overwrite = TRUE)
```

Or create only the workflow you need:

```{r one-action}
useGithubAction(dbms_type = "postgresql", overwrite = TRUE)
useGithubAction(dbms_type = "sqlserver", overwrite = TRUE)
useGithubAction(dbms_type = "databricks", overwrite = TRUE)
```

The function writes workflow files to `.github/workflows`. Commit those files
to your repository so GitHub Actions can run them.

## Add the expected test files

The workflow templates run fixed test file names. Your package must provide the
file for each backend you enable:

| Backend | Required test file |
| --- | --- |
| PostgreSQL | `tests/testthat/test-postgresql.R` |
| SQL Server | `tests/testthat/test-sqlserver.R` |
| Databricks/Spark | `tests/testthat/test-databricks.R` |

Each test file should call your package code, create the test CDM with
`patientsCDM(dbms = ...)`, run the study logic, and use `testthat` expectations
to check the result. If a test has an error or a failed expectation, the action
is intended to fail.

## PostgreSQL and SQL Server actions

For PostgreSQL and SQL Server, the GitHub Actions workflow creates a Docker
container inside the action run. That means users do not need to configure
database secrets for these two workflows.

The action sets the connection environment variables itself and tests against
the temporary container. This makes these two workflows useful for routine CI,
because each run starts from a fresh database service.

## Databricks action

For Databricks, the action connects to your own remote Databricks workspace.
You must configure the required repository secrets in GitHub:

| GitHub secret | Purpose |
| --- | --- |
| `DATABRICKS_HOST` | URL of the Databricks workspace |
| `DATABRICKS_TOKEN` | Token used to connect |
| `DATABRICKS_HTTPPATH` | SQL warehouse or cluster HTTP path |

Optional Databricks secrets are:

| GitHub secret | Default when not set |
| --- | --- |
| `DATABRICKS_WORKSPACE` | `hive_metastore` |
| `DATABRICKS_ODBC_DRIVER_URL` | The default Simba Spark ODBC driver URL used by the workflow |

Because Databricks uses a real remote workspace, make sure the token has enough
permission to create schemas and tables, insert the test data, query the result,
and remove the test schema afterwards.

## Suggested CI workflow

Start with PostgreSQL or SQL Server first, because those workflows create their
own database containers and do not require external credentials. Add Databricks
after you have confirmed that the remote workspace, token, and HTTP path work
from GitHub Actions.

Keep the backend-specific tests small and deterministic. A good CI test should
load a small Unit Test Definition, run one important part of your package logic,
and assert the expected counts, dates, or cohort records.

For release preparation, these backend-specific workflows can also act as a
quality gate. In `darwin-eu-dev/TestGenerator`, the `CRAN-release-check`
workflow only continues when the same `develop` commit already has successful
runs for `R-CMD-check`, `R Package Tests (PostgreSQL)`,
`R Package Tests (Spark Databricks)`, and `R Package Tests (SQL Server)`.
This means the CRAN checks are run on a commit that has already been tested
against the supported database backends.

---
title: "Getting started with TestGenerator"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with TestGenerator}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

TestGenerator helps you test pharmacoepidemiological study code against a
small, explicit OMOP CDM test population. The typical workflow is:

1. Create a small patient dataset in Excel or CSV files.
2. Convert that dataset to a Unit Test Definition JSON file.
3. Load the JSON into a blank CDM.
4. Run your study code and assert the expected results.

This vignette uses the ICU sample population included with the package.

## Create a Unit Test Definition

An Excel input file should contain one sheet per OMOP CDM table. For example,
the sheet names can include `person`, `observation_period`, `visit_occurrence`,
`condition_occurrence`, `drug_exposure`, and `measurement`.

```{r create-json}
library(TestGenerator)

file_path <- system.file(
  "extdata",
  "icu_sample_population.xlsx",
  package = "TestGenerator"
)

output_path <- file.path(tempdir(), "testgenerator-example")
dir.create(output_path, showWarnings = FALSE, recursive = TRUE)

readPatients(
  filePath = file_path,
  testName = "icu_sample",
  outputPath = output_path,
  cdmVersion = "5.4"
)
```

This writes `icu_sample.json` to `output_path`. Keeping these JSON files in
`tests/testthat/testCases` makes them easy to reuse from package tests. When
`outputPath = NULL`, TestGenerator writes to that default test case folder.

## Load the Test Population into a CDM

Use `patientsCDM()` to create a CDM reference containing the small patient
population and a complete vocabulary. By default, the CDM is created in DuckDB.

```{r load-cdm}
cdm <- patientsCDM(
  pathJson = output_path,
  testName = "icu_sample",
  cdmVersion = "5.4"
)

cdm[["person"]]
```

If `pathJson = NULL`, TestGenerator looks for JSON files in
`tests/testthat/testCases`.

```{r default-test-path}
cdm <- patientsCDM(
  pathJson = NULL,
  testName = "icu_sample",
  cdmVersion = "5.4"
)
```

## Use the CDM in Unit Tests

Once the test CDM is available, run the same study code you use on a real CDM.
The package includes example cohort definitions under `inst/extdata/test_cohorts`.

```{r cohort-test}
library(CDMConnector)
library(dplyr)
library(testthat)

test_cohorts <- system.file(
  "extdata",
  "test_cohorts",
  package = "TestGenerator"
)

cohort_set <- readCohortSet(test_cohorts)

cdm <- generateCohortSet(
  cdm = cdm,
  cohortSet = cohort_set,
  name = "test_cohorts"
)

cohort_attrition <- attrition(cdm[["test_cohorts"]])

excluded_records <- cohort_attrition |>
  pull(excluded_records) |>
  sum()

expect_equal(excluded_records, 0)
```

In a package test, place this code in `tests/testthat/test-*.R` and assert the
specific counts, dates, durations, or intersections that your study should
produce for the micro population.

## Start from a Blank Excel Template

If you want to design a new test population from scratch, create an Excel
workbook with the required CDM table columns.

```{r generate-template}
generateTestTables(
  tableNames = c(
    "person",
    "observation_period",
    "visit_occurrence",
    "condition_occurrence",
    "drug_exposure",
    "measurement"
  ),
  cdmVersion = "5.4",
  outputFolder = output_path,
  filename = "my_test_population"
)
```

Fill in the workbook rows for the small set of patients needed by your test,
then pass the completed workbook to `readPatients()`.

## CSV Inputs

For CSV inputs, place one file per CDM table in a folder. File names should
match the table names, for example `person.csv` and `observation_period.csv`.

```{r csv-input}
csv_path <- system.file(
  "extdata",
  "mimic_sample",
  package = "TestGenerator"
)

readPatients.csv(
  filePath = csv_path,
  testName = "mimic_sample",
  outputPath = output_path,
  cdmVersion = "5.4"
)
```

For source datasets with very large integer identifiers, set
`reduceLargeIds = TRUE`.

## Clean Up

For local DuckDB examples, disconnect when the test has finished.

```{r cleanup}
DBI::dbDisconnect(CDMConnector::cdmCon(cdm), shutdown = TRUE)
unlink(output_path, recursive = TRUE)
```

---
title: "Remote databases"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Remote databases}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

DuckDB is the default backend in TestGenerator, and it is usually the fastest
place to start. Remote database testing is useful when you also want to check
that your study code works on another SQL engine. This is especially helpful
for code that will run in PostgreSQL, SQL Server, or Databricks/Spark in
production.

The idea is the same as with local testing: you create a small, controlled test
population, load it into a CDM, run your study code, and assert the expected
result. The difference is that `patientsCDM()` uploads the test CDM to a remote
database and returns a CDM reference connected to that backend.

## Use `patientsCDM()` with a remote database

First create a Unit Test Definition JSON file with `readPatients()` or
`readPatients.csv()`. Then call `patientsCDM()` with the backend you want to
test.

```{r postgresql-example}
library(TestGenerator)

cdm <- patientsCDM(
  pathJson = "tests/testthat/testCases",
  testName = "my_test_population",
  cdmVersion = "5.4",
  dbms = "postgresql"
)
```

The supported remote values for `dbms` are:

| Backend | `dbms` value |
| --- | --- |
| PostgreSQL | `"postgresql"` |
| SQL Server | `"sqlserver"` |
| Databricks/Spark | `"databricks"` |

For remote databases, `patientsCDM()` creates the CDM locally first, trims the
vocabulary to what the test population needs, uploads the result to the remote
database, and returns the remote CDM reference. By default, TestGenerator creates
a temporary test schema for the upload. You can also provide `writeSchema` if
you want to choose the schema name yourself.

```{r write-schema}
cdm <- patientsCDM(
  pathJson = "tests/testthat/testCases",
  testName = "my_test_population",
  cdmVersion = "5.4",
  dbms = "sqlserver",
  writeSchema = "testgenerator_my_case"
)
```

When the test has finished, clean up the remote schema and close the connection:

```{r cleanup-remote}
cleanupTestCdm(cdm)
```

This is important for shared remote databases. The test schemas are small, but
cleaning them up keeps the database tidy and avoids name conflicts in later
runs.

## Environment variables for direct use

When you call `patientsCDM()` directly against a remote database, TestGenerator
reads the connection details from environment variables.

| Backend | Required environment variables |
| --- | --- |
| PostgreSQL | `POSTGRESQL_SERVER`, `POSTGRESQL_DBNAME`, `POSTGRESQL_PORT`, `POSTGRESQL_USER`, `POSTGRESQL_PASSWORD` |
| SQL Server | `SQLSERVER_SERVER`, `SQLSERVER_DBNAME`, `SQLSERVER_PORT`, `SQLSERVER_USER`, `SQLSERVER_PASSWORD` |
| Databricks/Spark | `DATABRICKS_HOST`, `DATABRICKS_TOKEN`, `DATABRICKS_HTTPPATH` |

For Databricks, TestGenerator also reads `DATABRICKS_USER` and
`DATABRICKS_WORKSPACE` when they are set. If they are not set, it uses `token`
as the user and `hive_metastore` as the workspace/catalog.

For SQL Server, TestGenerator reads `SQL_SERVER_DRIVER` when it is set. If it is
not set, it uses `ODBC Driver 18 for SQL Server`.

A typical local `.Renviron` setup could look like this:

```{bash envvars, eval=FALSE}
POSTGRESQL_SERVER=localhost
POSTGRESQL_DBNAME=cdm
POSTGRESQL_PORT=5432
POSTGRESQL_USER=postgres
POSTGRESQL_PASSWORD=your-password

SQLSERVER_SERVER=localhost
SQLSERVER_DBNAME=cdm
SQLSERVER_PORT=1433
SQLSERVER_USER=sa
SQLSERVER_PASSWORD=your-password

DATABRICKS_HOST=https://your-workspace.cloud.databricks.com
DATABRICKS_TOKEN=your-token
DATABRICKS_HTTPPATH=/sql/1.0/warehouses/your-warehouse
```

Use values that match your own database or Databricks workspace. The database
user needs permission to create schemas, create tables, insert data, read data,
and drop the test schema during cleanup.

## A minimal backend test pattern

A backend-specific test usually follows this shape:

```{r test-pattern}
testthat::test_that("study logic works on PostgreSQL", {
  cdm <- NULL
  on.exit({
    if (!is.null(cdm)) {
      TestGenerator::cleanupTestCdm(cdm)
    }
  }, add = TRUE)

  cdm <- TestGenerator::patientsCDM(
    pathJson = "tests/testthat/testCases",
    testName = "my_test_population",
    cdmVersion = "5.4",
    dbms = "postgresql"
  )

  result <- myPackage::runMyStudy(cdm)

  testthat::expect_equal(result$n_subjects, 3)
})
```

The exact expectations should be specific to your micro population. Good tests
usually check counts, dates, cohort entry and exit, exclusions, or any other
result that should be predictable from the small input dataset.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download.R
\name{checkRemoteFileAvailable}
\alias{checkRemoteFileAvailable}
\title{Check if a given remote file is available for download}
\usage{
checkRemoteFileAvailable(remoteFile)
}
\arguments{
\item{remoteFile}{a remote resource}
}
\value{
NULL if the remote resource is not available, other "success"
}
\description{
Check if a given remote file is available for download
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/patients.R
\name{patientsCDM}
\alias{patientsCDM}
\title{Pushes test population into a blank CDM.}
\usage{
patientsCDM(
  pathJson = NULL,
  testName = NULL,
  cdmVersion = "5.4",
  cdmName = NULL,
  dbms = "duckdb",
  writeSchema = NULL
)
}
\arguments{
\item{pathJson}{Directory where the sample populations in json are located. If NULL, gets the default inst/testCases directory.}

\item{testName}{Name of the sample population JSON file. If NULL it will push the first sample population in the testCases directory.}

\item{cdmVersion}{cdm version, default "5.4".}

\item{cdmName}{Name of the cdm, default NULL.}

\item{dbms}{Database management system to use. One of "duckdb", "databricks",
"sqlserver" or "postgresql". Default is "duckdb" which creates a local
DuckDB CDM. For remote databases the function creates the CDM locally,
trims the vocabulary, uploads to a new schema on the remote database,
and returns the remote CDM reference. Remote databases require
environment variables to be set. Call \code{usethis::edit_r_environ()}
to set them.}

\item{writeSchema}{Optional schema name to use on the remote database.
If NULL (default), a unique schema is created automatically. Only used
when \code{dbms} is not "duckdb".}
}
\value{
A CDM reference object with a sample population.
}
\description{
Pushes test population into a blank CDM.
}
\examples{
\donttest{
filePath <- system.file("extdata", "testPatientsRSV.xlsx", package = "TestGenerator")
TestGenerator::readPatients(filePath = filePath, outputPath = tempdir())
cdm <- TestGenerator::patientsCDM(pathJson = tempdir(), testName = "test")
duckdb::duckdb_shutdown(duckdb::duckdb())
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/patients.R
\name{readPatients.csv}
\alias{readPatients.csv}
\title{Converts a sample of patients in CSV format into a Unit Testing Definition JSON file.}
\usage{
readPatients.csv(
  filePath = NULL,
  testName = "test",
  outputPath = NULL,
  cdmVersion = "5.4",
  reduceLargeIds = FALSE
)
}
\arguments{
\item{filePath}{Path to the test patient data in CSV format. Multiple CSV files representing tables tables from the OMOP-CDM must be provided, e.g. person.csv, drug_exposure.csv, condition_ocurrence.csv, etc.}

\item{testName}{Name for the test population file in character.}

\item{outputPath}{Path of the output file, if NULL, a folder will be created in the project folder inst/testCases.}

\item{cdmVersion}{cdm version, default "5.4".}

\item{reduceLargeIds}{Reduces the length of very long ids generally in int64 format, such as those found in the MIMIC-IV database.}
}
\value{
A JSON file with sample patients inside the project directory.
}
\description{
Converts a sample of patients in CSV format into a Unit Testing Definition JSON file.
}
\examples{
filePath <- system.file("extdata", "mimic_sample", package = "TestGenerator")
readPatients.csv(
  filePath = filePath,
  outputPath = tempdir(),
  cdmVersion = "5.3"
  )

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/useGithubAction.R
\name{useGithubAction}
\alias{useGithubAction}
\title{Create GitHub Action workflows for specific DBMS}
\usage{
useGithubAction(
  dbms_type = c("postgresql", "sqlserver", "databricks"),
  overwrite = FALSE
)
}
\arguments{
\item{dbms_type}{A character vector of supported DBMS types to process.
Supported values are \code{"postgresql"}, \code{"sqlserver"}, and
\code{"databricks"}.}

\item{overwrite}{A logical value indicating whether to overwrite existing
workflow files in the \code{.github/workflows} directory. Defaults to
\code{FALSE}.}
}
\value{
Invisibly returns \code{TRUE} when the selected workflow files are
copied successfully.
}
\description{
Copies GitHub Action workflow templates from the package's
\code{inst/workflows} directory to the \code{.github/workflows} directory of
the current project. The copied workflows can be used to run backend-specific
tests for PostgreSQL, SQL Server, and Databricks.
}
\examples{
\dontrun{
useGithubAction(dbms_type = "postgresql")
useGithubAction(
  dbms_type = c("postgresql", "sqlserver", "databricks"),
  overwrite = TRUE
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download.R
\name{downloadTestData}
\alias{downloadTestData}
\title{Download Test Data Files}
\usage{
downloadTestData(
  datasetName = "mimicIV",
  cdmVersion = "5.3",
  pathToData = Sys.getenv("STUDY_DATASETS"),
  overwrite = FALSE
)
}
\arguments{
\item{datasetName}{The data set name as found on https://github.com/darwin-eu/EunomiaDatasets. The
data set name corresponds to the folder with the data set ZIP files}

\item{cdmVersion}{The OMOP CDM version. This version will appear in the suffix of the data file,
for example: synpuf_5.3.zip. Default: '5.3'}

\item{pathToData}{The path where the Eunomia data is stored on the file system., By default the
value of the environment variable "EUNOMIA_DATA_FOLDER" is used.}

\item{overwrite}{Control whether the existing archive file will be overwritten should it already exist.}
}
\value{
Invisibly returns the destination if the download was successful.
}
\description{
Download Test Data Files
}
\examples{
\donttest{
downloadTestData(pathToData = tempdir())
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/patients.R
\name{checkTablesColumns}
\alias{checkTablesColumns}
\title{Check if the given tables and columns are valid and return the loaded data.}
\usage{
checkTablesColumns(cdmVersion, filePath, extraTable = FALSE)
}
\arguments{
\item{cdmVersion}{cdm version}

\item{filePath}{Path to the test patient data in xlsx format}

\item{extraTable}{if extra tables are provided or not, default FALSE}
}
\value{
a named list containing the loaded data
}
\description{
Check if the given tables and columns are valid and return the loaded data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualisations.R
\name{graphCohort}
\alias{graphCohort}
\title{Deprecated cohort timeline visualisation}
\usage{
graphCohort(subject_id, cohorts = list())
}
\arguments{
\item{subject_id}{Deprecated.}

\item{cohorts}{Deprecated.}
}
\value{
Invisibly returns \code{NULL}.
}
\description{
\code{graphCohort()} is deprecated and will be removed in a future release.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/patients.R
\name{readPatients}
\alias{readPatients}
\title{Converts a sample of patients into Unit Testing Definition JSON file.}
\usage{
readPatients(
  filePath = NULL,
  testName = "test",
  outputPath = NULL,
  cdmVersion = "5.4",
  extraTable = FALSE
)
}
\arguments{
\item{filePath}{Path to the test patient data in Excel format. The Excel has sheets that represent tables from the OMOP-CDM, e.g. person, drug_exposure, condition_ocurrence, etc.}

\item{testName}{A name of the test population in character.}

\item{outputPath}{Path of the output file, if NULL, a folder will be created in the project folder inst/testCases.}

\item{cdmVersion}{cdm version, default "5.4".}

\item{extraTable}{Name of non-standard tables to be included in the test CDM.}
}
\value{
A JSON file with sample patients inside the project directory.
}
\description{
Converts a sample of patients into Unit Testing Definition JSON file.
}
\examples{
filePath <- system.file("extdata", "testPatientsRSV.xlsx", package = "TestGenerator")
readPatients(filePath = filePath, outputPath = tempdir())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generateTestTables.R
\name{generateTestTables}
\alias{generateTestTables}
\title{Generates an Excel file with sheets that correspond to an OMOP-CDM tables.}
\usage{
generateTestTables(
  tableNames = c("person", "observation_period", "visit_occurrence", "visit_detail",
    "condition_occurrence", "drug_exposure", "procedure_occurrence", "measurement",
    "observation", "death", "drug_era", "condition_era", "dose_era", "location",
    "care_site", "provider"),
  cdmVersion,
  outputFolder,
  filename = paste0("test_cdm_", cdmVersion)
)
}
\arguments{
\item{tableNames}{A list specifying the table names to include in the Excel file.}

\item{cdmVersion}{The CDM version to use for creating the requested tables (either 5.3 or 5.4).}

\item{outputFolder}{The folder where the Excel file will be saved.}

\item{filename}{The name of the Excel file. It does not include the extension .xlsx.}
}
\value{
An Excel file with the tables requested.
}
\description{
Generates an Excel file with sheets that correspond to an OMOP-CDM tables.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/patientsRemote.R
\name{cleanupTestCdm}
\alias{cleanupTestCdm}
\title{Clean up a test CDM on a remote database}
\usage{
cleanupTestCdm(cdm)
}
\arguments{
\item{cdm}{A CDM reference object created by \code{patientsCDM()} with a
remote database backend.}
}
\value{
Invisibly returns NULL.
}
\description{
Drops the schema containing the test CDM and disconnects from the database.
}
